package org.byteam.dfa;

import android.Manifest;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.net.Uri;
import android.os.Build;
import android.os.Environment;
import android.util.Log;

import androidx.core.app.ActivityCompat;

import org.byteam.dfa.model.InstallResult;
import org.byteam.dfa.model.Patch;
import org.byteam.dfa.util.UriUtils;
import org.byteam.dfa.util.Utils;

import org.json.JSONObject;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.PluginRegistry;

public class DfaDelegate implements PluginRegistry.ActivityResultListener, PluginRegistry.RequestPermissionsResultListener {
    private static final String TAG = "DfaDelegate";
    private static final int REQUEST_CODE = 1008;

    private final Activity activity;
    private MethodChannel.Result pendingResult;

    DfaDelegate(Activity activity) {
        this.activity = activity;
    }

    @Override
    public boolean onActivityResult(int requestCode, int resultCode, final Intent data) {
        if (requestCode == REQUEST_CODE && resultCode == Activity.RESULT_OK) {
            if (data.getData() != null) {
                new Thread(new Runnable() {
                    @Override
                    public void run() {
                        final Uri uri = data.getData();
                        Log.i(TAG, "File URI:" + uri.toString());
                        String fullPath = UriUtils.getPath(uri, activity);

                        if (fullPath == null) {
                            fullPath = UriUtils.getPath2(activity, uri);
                        }

                        if (fullPath != null) {
                            Log.i(TAG, "File path: " + fullPath);
                            finishWithSuccess(fullPath);
                        } else {
                            finishWithError("unknown_path", "Failed to retrieve path.");
                        }
                    }
                }).start();
            } else {
                finishWithError("null_data", "Unknown result data.");
            }
            return true;
        } else if (requestCode == REQUEST_CODE && resultCode == Activity.RESULT_CANCELED) {
            Log.i(TAG, "User cancelled the request");
            finishWithSuccess(null);
            return true;
        } else if (requestCode == REQUEST_CODE) {
            finishWithError("unknown_activity", "Unknown activity error.");
        }
        return false;
    }

    @Override
    public boolean onRequestPermissionsResult(int requestCode, String[] permissions, int[] grantResults) {
        if (requestCode != REQUEST_CODE) {
            return false;
        }
        boolean permissionGranted = false;
        for (int i = 0, len = grantResults.length; i < len; i++) {
            if (Manifest.permission.READ_EXTERNAL_STORAGE.equals(permissions[i])) {
                permissionGranted = grantResults[i] == PackageManager.PERMISSION_GRANTED;
                break;
            }
        }
        if (permissionGranted) {
            choosePatchFile();
        } else {
            finishWithError("permission_denied", "User did not allow reading external storage");
        }
        return true;
    }

    /**
     * Choose a patch file.
     */
    void choosePatchFile(MethodChannel.Result result) {
        if (!checkPendingResult(result)) {
            finishWithAlreadyActiveError(result);
            return;
        }
        if (ActivityCompat.checkSelfPermission(activity, Manifest.permission.READ_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(activity, new String[]{Manifest.permission.READ_EXTERNAL_STORAGE}, REQUEST_CODE);
            return;
        }
        choosePatchFile();
    }

    private void choosePatchFile() {
        Intent intent = new Intent(Intent.ACTION_GET_CONTENT);
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
            intent.setAction(Intent.ACTION_OPEN_DOCUMENT);
        }
        final Uri uri = Uri.parse(Environment.getExternalStorageDirectory().getPath() + File.separator);
        intent.setDataAndType(uri, "application/zip"); // 过滤出zip类型的文件
        if (intent.resolveActivity(activity.getPackageManager()) != null) {
            activity.startActivityForResult(intent, REQUEST_CODE);
        } else {
            Log.e(TAG, "Can't find a valid activity to handle the request. Make sure you've a file explorer installed.");
            finishWithError("invalid_format_type", "Can't handle the provided file type.");
        }
    }

    private void finishWithSuccess(final Object data) {
        if (this.pendingResult != null) {
            this.pendingResult.success(data);
            this.clearPendingResult();
        }
    }

    private void finishWithError(final String errorCode, final String errorMessage) {
        if (this.pendingResult == null) {
            return;
        }
        this.pendingResult.error(errorCode, errorMessage, null);
        this.clearPendingResult();
    }

    private void finishWithAlreadyActiveError(final MethodChannel.Result result) {
        result.error("already_active", "Original result is already active", null);
    }

    private boolean checkPendingResult(final MethodChannel.Result result) {
        if (this.pendingResult != null) {
            return false;
        }
        this.pendingResult = result;
        return true;
    }

    private void clearPendingResult() {
        this.pendingResult = null;
    }

    /**
     * 安装patch
     *
     * @param path   patch absolute path
     * @param result result
     */
    void installPatch(Context applicationContext, String path, final MethodChannel.Result result) {
        if (!checkPendingResult(result)) {
            finishWithAlreadyActiveError(result);
            return;
        }
        if (!Utils.isAOT(applicationContext)) {
            finishWithSuccess(InstallResult.error("Non-AOT mode"));
            return;
        }
        if (path == null || path.isEmpty()) {
            finishWithSuccess(InstallResult.error("Invalid patch path(null or empty)"));
            return;
        }
        File patchFile = new File(path);
        if (!patchFile.exists()) {
            finishWithSuccess(InstallResult.error("File not exists for given path: " + path));
            return;
        }
        Patch patch;
        try {
            patch = getArchiveInfo(patchFile);
        } catch (Exception e) {
            e.printStackTrace();
            finishWithSuccess(InstallResult.error(e.getMessage()));
            return;
        }
        if (patch == null) {
            finishWithSuccess(InstallResult.error("Empty content"));
            return;
        }
        if (patch.getFiles() == null || patch.getFiles().isEmpty()) {
            finishWithSuccess(InstallResult.error("Patch contains empty files"));
            return;
        }
        Patch existPatch = DfaLoader.getPatch(applicationContext);
        if (existPatch != null && existPatch.getVersion().equals(patch.getVersion())) {
            finishWithSuccess(InstallResult.error(String.format(
                    "Install patch(%s) failed, there is a patch installed with the same version.",
                    patch.getInstallPath())));
            return;
        }
        if (Utils.getAppVersionCode(applicationContext) != patch.getAppVersionCode()) {
            finishWithSuccess(InstallResult.error("Patch version does not match the app version"));
            return;
        }

        Extractor.ExtractTask extractTask = new Extractor.ExtractTask(
                applicationContext, patchFile, patch, new Extractor.Callback() {
            @Override
            public void result(Throwable t) {
                if (t == null) {
                    finishWithSuccess(InstallResult.success());
                } else {
                    finishWithSuccess(InstallResult.error(t.getMessage()));
                }
            }
        });
        extractTask.execute();
    }

    /**
     * Get patch archive info.
     */
    private Patch getArchiveInfo(File patchFile) throws Exception {
        ZipFile zipFile = null;
        try {
            zipFile = new ZipFile(patchFile);
            ZipEntry manifest = zipFile.getEntry(DfaLoader.MANIFEST_FILE_NAME);
            if (manifest == null || manifest.isDirectory()) {
                throw new FileNotFoundException("Invalid archive file for missing " + DfaLoader.MANIFEST_FILE_NAME);
            }
            InputStream is = zipFile.getInputStream(manifest);
            String content = Utils.readInputStreamAsString(is);
            if (content == null) {
                return null;
            }
            JSONObject json = new JSONObject(content);
            return new Patch(json);
        } finally {
            try {
                if (zipFile != null) {
                    zipFile.close();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    void cleanAll() {
        File dataDir = DfaLoader.getDataDirectory(activity);
        if (dataDir != null && dataDir.exists()) {
            Utils.deleteFile(dataDir);
        }
        DfaLoader.getSharedPreferences(activity).edit().clear().apply();
    }
}
