package org.byteam.dfa;

import android.content.Context;
import android.content.SharedPreferences;
import android.util.Log;

import androidx.annotation.Keep;

import org.byteam.dfa.model.Patch;
import org.byteam.dfa.util.Utils;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.File;
import java.lang.reflect.Field;

import io.flutter.embedding.engine.loader.FlutterLoader;

public class DfaLoader {
    private static final String TAG = "DfaLoader";

    static final String MANIFEST_FILE_NAME = "manifest.json";
    static final String KEY_LOADED_PATCH = "loaded_patch";
    static final String KEY_INSTALLED_PATCH = "installed_patch";

    @Keep
    public static void load(Context context) {
        try {
            if (!Utils.isAOT(context)) { // ignore debug mode
                Log.i(TAG, "load: ignore Non-AOT mode");
                return;
            }
            Patch patch = getPatch(context);
            if (patch == null || patch.getInstallPath() == null) {
                Log.i(TAG, "load: no patch");
                return;
            }

            File patchDir = new File(patch.getInstallPath());
            if (!patchDir.exists()) {
                Log.e(TAG, String.format("%s not exists.", patchDir.getAbsolutePath()));
                return;
            }

            File libDir = new File(patchDir, "lib");
            if (!libDir.exists()) {
                Log.i(TAG, libDir.getAbsolutePath() + " not exists.");
                return;
            }
            File abiDir = new File(patchDir, "lib/" + Utils.getCurrentAbi(context));
            if (!abiDir.exists()) {
                Log.e(TAG, abiDir.getAbsolutePath() + " not exists.");
                return;
            }
            File appFile = new File(abiDir, "libapp.so");
            if (!appFile.exists()) {
                Log.e(TAG, appFile.getAbsolutePath() + " not exists.");
                return;
            }
            // inject
            String libPath = appFile.getAbsolutePath();
            FlutterLoader flutterLoader = FlutterLoader.getInstance();
            Field field = FlutterLoader.class.getDeclaredField("aotSharedLibraryName");
            field.setAccessible(true);
            field.set(flutterLoader, libPath);
            // update sp
            SharedPreferences sp = getSharedPreferences(context);
            SharedPreferences.Editor editor = sp.edit();
            if (sp.contains(KEY_INSTALLED_PATCH)) {
                editor.remove(KEY_INSTALLED_PATCH);
            }
            editor.putString(KEY_LOADED_PATCH, patch.toJsonString());
            editor.apply();
            // crash handler
            CrashHandler.getInstance().init(context);
            Log.i(TAG, "load: success.");
        } catch (Exception e) {
            Log.e(TAG, "load: failed", e);
        }
    }

    public static SharedPreferences getSharedPreferences(Context context) {
        return context.getSharedPreferences("dfa", Context.MODE_PRIVATE);
    }

    public static File getDataDirectory(Context context) {
        return context.getDir("dfa", Context.MODE_PRIVATE);
    }

    static boolean deletePatch(Patch patch) {
        String path = patch.getInstallPath();
        File file = new File(path);
        if (file.exists()) {
            return Utils.deleteFile(file);
        }
        return false;
    }

    /**
     * 优先查找已安装的，再查找已加载的
     */
    static Patch getPatch(Context context) {
        Patch patch = getInstalledPatch(context);
        if (patch == null) {
            patch = getLoadedPatch(context);
        }
        if (patch != null) {
            if (patch.getAppVersionCode() != Utils.getAppVersionCode(context)) {
                deletePatch(patch);
                return null;
            }
        }
        return patch;
    }

    static Patch getInstalledPatch(Context context) {
        SharedPreferences sp = getSharedPreferences(context);
        String content = sp.getString(KEY_INSTALLED_PATCH, null);
        if (content == null || content.trim().isEmpty()) {
            return null;
        }
        try {
            JSONObject jsonObject = new JSONObject(content);
            Patch patch = new Patch(jsonObject);
            if (!new File(patch.getInstallPath()).exists()) {
                sp.edit().remove(KEY_INSTALLED_PATCH).apply();
                return null;
            }
            return patch;
        } catch (JSONException e) {
            e.printStackTrace();
        }
        return null;
    }

    static Patch getLoadedPatch(Context context) {
        SharedPreferences sp = getSharedPreferences(context);
        String content = sp.getString(KEY_LOADED_PATCH, null);
        if (content == null || content.trim().isEmpty()) {
            return null;
        }
        try {
            JSONObject jsonObject = new JSONObject(content);
            Patch patch = new Patch(jsonObject);
            if (!new File(patch.getInstallPath()).exists()) {
                sp.edit().remove(KEY_LOADED_PATCH).apply();
                return null;
            }
            return patch;
        } catch (JSONException e) {
            e.printStackTrace();
        }
        return null;
    }
}
