package org.byteam.dfa;

import android.annotation.SuppressLint;
import android.content.Context;
import android.content.SharedPreferences;
import android.os.AsyncTask;
import android.util.Log;

import org.byteam.dfa.model.Asset;
import org.byteam.dfa.model.Patch;
import org.byteam.dfa.util.BSPatch;
import org.byteam.dfa.util.Utils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Locale;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

@SuppressLint("StaticFieldLeak")
class Extractor {
    private static final String TAG = "Dfa.Extractor";

    public interface Callback {
        void result(Throwable t);
    }

    public static class ExtractTask extends AsyncTask<Void, Void, Throwable> {
        // Application context
        private Context context;
        // patch 文件
        private File patchFile;
        // patch 信息
        private Patch patch;
        // 回调
        private Callback callback;

        public ExtractTask(Context context, File patchFile, Patch patch, Callback callback) {
            this.context = context;
            this.patchFile = patchFile;
            this.patch = patch;
            this.callback = callback;
        }

        @Override
        protected Throwable doInBackground(Void... voids) {
            ZipFile apkArchive = null;
            ZipFile patchArchive = null;
            try {
                // patch安装目录的文件夹名字
                String installDirName = String.format(Locale.getDefault(), "dfa_%d_%s",
                        patch.getAppVersionCode(),
                        patch.getVersion());
                File patchInstallDir = new File(DfaLoader.getDataDirectory(context), installDirName);
                patchInstallDir.mkdirs();

                // diff/extract
                String abi = Utils.getCurrentAbi(context);
                apkArchive = new ZipFile(context.getPackageManager().getApplicationInfo(
                        context.getPackageName(), 0).publicSourceDir);
                patchArchive = new ZipFile(patchFile);
                for (Asset asset : patch.getFiles()) {
                    String name = asset.getName();
                    Log.d(TAG, String.format("processing file: %s", name));
                    if (name.startsWith("lib/") && !name.startsWith("lib/" + abi + "/")) {
                        Log.i(TAG, String.format("ignore %s for mismatching abi(%s).", name, abi));
                        continue;
                    }
                    ZipEntry patchEntry = patchArchive.getEntry(name);
                    if (patchEntry == null) {
                        throw new FileNotFoundException(
                                String.format("%s not found in patch.", name));
                    }
                    if (asset.isDiff()) {
                        ZipEntry baseEntry = apkArchive.getEntry(name);
                        if (baseEntry == null) {
                            throw new FileNotFoundException(
                                    String.format("%s not found in apk.", name));
                        }
                        long begin = System.currentTimeMillis();
                        byte[] oldData = Utils.readInputStream(apkArchive.getInputStream(baseEntry));
                        byte[] diffData = Utils.readInputStream(patchArchive.getInputStream(patchEntry));
                        if (oldData == null || diffData == null) {
                            throw new IOException("read data from zip entry failed.");
                        }
                        byte[] newData = BSPatch.bspatch(oldData, diffData);
                        String md5 = Utils.getMD5(newData);
                        if (!asset.getMd5().equalsIgnoreCase(md5)) {
                            throw new IllegalStateException(String.format("%s's md5 verify failed(%s != %s).", name, md5, asset.getMd5()));
                        }
                        long cost = System.currentTimeMillis() - begin;
                        Log.i(TAG, String.format("%s: bspatch cost %d ms", name, cost));
                        File targetFile = new File(patchInstallDir, name);
                        targetFile.getParentFile().mkdirs();
                        Utils.writeByteArrayToFile(targetFile, newData);
                    } else {
                        byte[] bytes = Utils.readInputStream(patchArchive.getInputStream(patchEntry));
                        File targetFile = new File(patchInstallDir, name);
                        targetFile.getParentFile().mkdirs();
                        Utils.writeByteArrayToFile(targetFile, bytes);
                    }
                }

                // 备份原始patch包
                // String dstPatchFileName = installDirName + ".zip";
                // File dstPatchFile = new File(PathUtils.getDataDirectory(context), dstPatchFileName);
                // Utils.copyFile(patchFile, dstPatchFile);

                // update fields
                patch.setInstallPath(patchInstallDir.getAbsolutePath());
                patch.setInstallTimestamp(System.currentTimeMillis());

                // update sp
                SharedPreferences sp = DfaLoader.getSharedPreferences(context);
                sp.edit().putString(DfaLoader.KEY_INSTALLED_PATCH, patch.toJsonString()).apply();
            } catch (Exception e) {
                e.printStackTrace();
                return e;
            } finally {
                try {
                    if (apkArchive != null) {
                        apkArchive.close();
                    }
                    if (patchArchive != null) {
                        patchArchive.close();
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            return null;
        }

        @Override
        protected void onPostExecute(Throwable t) {
            super.onPostExecute(t);
            if (callback != null) {
                callback.result(t);
            }
        }
    }
}
