package org.byteam.dfa.util;

import android.content.Context;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.os.Build;
import android.util.Log;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.zip.ZipFile;

public class Utils {
    private static final String TAG = "Dfa.Utils";

    private static long appVersionCode = 0L;
    private static String currentAbi;
    private static Boolean isAOT;

    public static long getAppVersionCode(Context context) {
        if (appVersionCode == 0) {
            try {
                PackageInfo pi = context.getPackageManager().getPackageInfo(context.getPackageName(), 0);
                if (Build.VERSION.SDK_INT >= 28) {
                    appVersionCode = pi.getLongVersionCode();
                } else {
                    appVersionCode = pi.versionCode;
                }
            } catch (PackageManager.NameNotFoundException e) {
                e.printStackTrace();
            }
        }

        return appVersionCode;
    }

    public static boolean isAOT(Context context) {
        if (isAOT == null) {
            retrieveDataFromApk(context);
        }
        return isAOT;
    }

    public static String getCurrentAbi(Context context) {
        if (currentAbi == null) {
            retrieveDataFromApk(context);
        }
        return currentAbi;
    }

    private static List<String> getSupportedAbis() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
            return Arrays.asList(Build.SUPPORTED_ABIS);
        } else {
            ArrayList<String> cpuAbis = new ArrayList<>(Arrays.asList(Build.CPU_ABI, Build.CPU_ABI2));
            cpuAbis.removeAll(Arrays.asList(null, ""));
            return cpuAbis;
        }
    }

    private static void retrieveDataFromApk(Context context) {
        List<String> supportedAbis = getSupportedAbis();
        ZipFile zipFile = null;
        try {
            // context.getApplicationInfo().nativeLibraryDir 返回的值类似为
            // /data/app/com.example-wFhOjQzNVJZtAz7n0W6-uA==/lib/arm，所以无法使用
            String apkPath = context.getApplicationInfo().publicSourceDir;
            zipFile = new ZipFile(apkPath);
            if (currentAbi == null) {
                for (String abi : supportedAbis) {
                    if (zipFile.getEntry("lib/" + abi + "/libflutter.so") != null) {
                        currentAbi = abi;
                        break;
                    }
                }
            }
            if (isAOT == null) {
                if (zipFile.getEntry("lib/" + currentAbi + "/libapp.so") != null) {
                    isAOT = true;
                } else {
                    isAOT = false;
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (currentAbi == null) {
                currentAbi = supportedAbis.get(0);
                Log.w(TAG, "use default abi " + currentAbi);
            }
            if (isAOT == null) {
                isAOT = true;
            }
            try {
                if (zipFile != null) {
                    zipFile.close();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public static byte[] readInputStream(InputStream is) throws IOException {
        if (is == null) {
            return null;
        }
        byte[] buf = new byte[1024 * 8];
        ByteArrayOutputStream baos = null;
        try {
            int num;
            baos = new ByteArrayOutputStream();
            while ((num = is.read(buf, 0, buf.length)) != -1) {
                baos.write(buf, 0, num);
            }
            return baos.toByteArray();
        } finally {
            is.close();
            if (baos != null) {
                baos.close();
            }
        }

    }

    public static String readInputStreamAsString(InputStream is) throws IOException {
        byte[] bytes = readInputStream(is);
        if (bytes != null) {
            return new String(bytes);
        }
        return null;
    }

    public static byte[] readFile(File file) throws IOException {
        InputStream is = null;
        try {
            is = new FileInputStream(file);
            return readInputStream(is);
        } finally {
            if (is != null) {
                is.close();
            }
        }
    }

    public static void writeByteArrayToFile(File file, byte[] data) throws IOException {
        BufferedOutputStream bos = null;
        try {
            bos = new BufferedOutputStream(new FileOutputStream(file));
            bos.write(data);
            bos.close();
        } finally {
            if (bos != null) {
                bos.close();
            }
        }
    }

    public static void writeStringToFile(File file, String content) throws IOException {
        writeByteArrayToFile(file, content.getBytes());
    }

    public static void copyFile(File src, File dst) throws IOException {
        InputStream in = null;
        OutputStream out = null;
        try {
            in = new FileInputStream(src);
            out = new FileOutputStream(dst);
            copy(in, out);
        } finally {
            if (in != null) {
                in.close();
            }
            if (out != null) {
                out.close();
            }
        }
    }

    /**
     * Remember to close the streams.
     */
    public static void copy(InputStream in, OutputStream out) throws IOException {
        final byte[] buffer = new byte[4 * 1024];
        int bytesRead;
        while ((bytesRead = in.read(buffer)) != -1) {
            out.write(buffer, 0, bytesRead);
        }
        out.flush();
    }

    public static boolean deleteFile(File file) {
        if (file.isFile()) {
            return file.delete();
        } else {
            File[] children = file.listFiles();
            if (children != null) {
                for (File f : children) {
                    if (!deleteFile(f)) {
                        return false;
                    }
                }
            }
            return file.delete();
        }
    }

    /**
     * 获取MD5，解决了左边缺'0'的问题
     */
    public static String getMD5(byte[] bytes) {
        MessageDigest digest;
        try {
            digest = MessageDigest.getInstance("MD5");
            digest.update(bytes);
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        byte[] result = digest.digest();
        StringBuilder sb = new StringBuilder();
        for (byte b : result) {
            String s = Integer.toHexString(b & 0xFF);
            if (s.length() == 1) {
                sb.append("0");
            }
            sb.append(s);
        }
        return sb.toString();
    }
}
