import 'dart:convert';
import 'dart:io';

import 'package:archive/archive.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

import 'asset_bundle.dart';
import 'dev_page.dart';
import 'install_result.dart';
import 'patch.dart';

class DfaManager {
  static const MethodChannel _channel = const MethodChannel('dfa');

  static const String _MANIFEST_FILE_NAME = 'manifest.json';

  DfaManager._();

  static AssetBundle getAssetBundle() => dfaAssetBundle;

  /// 打开开发者预览界面
  static openDevPage(BuildContext context) async {
    Navigator.push(context, MaterialPageRoute(builder: (context) => DevPage()));
  }

  /// 从文件选择器中选择一个patch压缩文件
  /// 返回file path，可能抛出[PlatformException]
  static Future<String> choosePatchFile() async {
    if (!Platform.isAndroid) {
      return null;
    }
    String result = await _channel.invokeMethod('choosePatchFile');
    return result;
  }

  /// 安装patch包
  static Future<InstallResult> installPatch(File patchFile) async {
    if (!Platform.isAndroid) {
      return null;
    }

    if (patchFile == null || !patchFile.existsSync()) {
      return null;
    }

    Map<String, dynamic> result =
        await _channel.invokeMapMethod('installPatch', {
      'path': patchFile.absolute.path,
    });
    return InstallResult.fromJson(result);
  }

  /// 获取已加载(当前正在使用)的 patch 信息
  static Future<Patch> getLoadedPatch() async {
    if (!Platform.isAndroid) {
      return null;
    }

    try {
      String result = await _channel.invokeMethod('loadedPatch');
      if (result == null) {
        return null;
      } else {
        Map<String, dynamic> json = jsonDecode(result);
        Patch patch = Patch.fromJson(json);
        return patch;
      }
    } catch (e) {
      print(e);
      return null;
    }
  }

  /// 获取已安装未加载的 patch 信息
  static Future<Patch> getInstalledPatch() async {
    if (!Platform.isAndroid) {
      return null;
    }

    try {
      String result = await _channel.invokeMethod('installedPatch');
      if (result == null) {
        return null;
      } else {
        Map<String, dynamic> json = jsonDecode(result);
        Patch patch = Patch.fromJson(json);
        return patch;
      }
    } catch (e) {
      print(e);
      return null;
    }
  }

  /// 清除所有 patch 相关的数据
  static Future<void> cleanAll() async {
    if (!Platform.isAndroid) {
      return;
    }

    await _channel.invokeMethod('cleanAll');
  }

  ///  获取 patch 压缩包的信息
  static Future<Patch> getArchiveInfo(File patchFile) async {
    if (patchFile == null || !patchFile.existsSync()) {
      return null;
    }

    Archive archive = ZipDecoder().decodeBytes(patchFile.readAsBytesSync());
    ArchiveFile manifest = archive.findFile(_MANIFEST_FILE_NAME);
    if (manifest == null || !manifest.isFile) {
      // invalid
      return null;
    }
    String content = utf8.decode(manifest.content);
    Map<String, dynamic> jsonContent = jsonDecode(content);
    Patch patch = Patch.fromJson(jsonContent);
    return patch;
  }

  /// 删除指定 patch 数据
  static Future<void> deletePatch(Patch patch) async {
    if (patch == null ||
        patch.installPath == null ||
        patch.installPath.isEmpty) {
      return;
    }
    Directory patchPath = Directory(patch.installPath);
    if (patchPath.existsSync()) {
      patchPath.deleteSync(recursive: true);
    }
  }
}
